import itertools
from enum import StrEnum
from functools import partial
from typing import Generator

import networkx as nx

from matchescu.similarity._matcher import Matcher
from matchescu.typing._references import EntityReference


class MatchEdgeType(StrEnum):
    MATCH = "match"
    POTENTIAL_MATCH = "potential_match"


class SimilarityGraph:
    def __init__(
        self, matcher: Matcher, max_non_match: float, min_match: float
    ) -> None:
        self.__g = nx.DiGraph()
        self.__matcher = matcher
        self.__max_bad = max_non_match
        self.__min_good = min_match
        self.__match_count = 0
        self.__potential_match_count = 0
        self.__non_match_count = 0

    def __repr__(self):
        return "SimilarityGraph(nodes={}, edges={}, match={}, non_match={}, maybe={})".format(
            len(self.__g.nodes),
            len(self.__g.edges),
            self.__match_count,
            self.__non_match_count,
            self.__potential_match_count,
        )

    @property
    def nodes(self):
        """Returns the nodes of the graph."""
        return self.__g.nodes

    @property
    def edges(self):
        """Returns the edges of the graph along with their similarity weights and types."""
        return self.__g.edges(data=True)

    @property
    def match_count(self):
        return self.__match_count

    @property
    def potential_match_count(self):
        return self.__potential_match_count

    @property
    def non_match_count(self):
        return self.__non_match_count

    def add(self, left: EntityReference, right: EntityReference) -> "SimilarityGraph":
        """Adds an edge between two entities based on similarity thresholds."""
        if left not in self.__g:
            self.__g.add_node(left)
        if right not in self.__g:
            self.__g.add_node(right)

        sim_score = self.__matcher(left, right)
        if sim_score >= self.__min_good:
            self.__g.add_edge(left, right, weight=sim_score, type=MatchEdgeType.MATCH)
            self.__match_count += 1
        elif self.__max_bad <= sim_score < self.__min_good:
            self.__g.add_edge(
                left, right, weight=sim_score, type=MatchEdgeType.POTENTIAL_MATCH
            )
            self.__potential_match_count += 1
        else:
            self.__non_match_count += 1
        return self

    @staticmethod
    def __has_expected_type(edge: tuple, edge_type: MatchEdgeType) -> bool:
        _, __, data = edge
        return data.get("type") == edge_type

    def edges_by_type(
        self, edge_type: MatchEdgeType
    ) -> Generator[tuple[EntityReference, EntityReference], None, None]:
        yield from itertools.starmap(
            lambda a, b, _: (a, b),
            filter(partial(self.__has_expected_type, edge_type=edge_type), self.edges),
        )

    def matches(self) -> Generator[tuple[EntityReference, EntityReference], None, None]:
        yield from self.edges_by_type(MatchEdgeType.MATCH)

    def potential_matches(
        self,
    ) -> Generator[tuple[EntityReference, EntityReference], None, None]:
        yield from self.edges_by_type(MatchEdgeType.POTENTIAL_MATCH)

    def is_match(self, left: EntityReference, right: EntityReference) -> bool:
        data = self.__g.get_edge_data(left, right, default={})
        return data.get("type") == MatchEdgeType.MATCH

    def is_potential_match(self, left: EntityReference, right: EntityReference) -> bool:
        data = self.__g.get_edge_data(left, right, default={})
        return data.get("type") == MatchEdgeType.POTENTIAL_MATCH

    def is_non_match(self, left: EntityReference, right: EntityReference) -> bool:
        return (left, right) not in self.__g.edges



from typing import Protocol

from matchescu.typing import EntityReference


class Matcher(Protocol):
    @property
    def non_match_threshold(self) -> float:
        """Similarity score (ranged 0..1) below which two references are considered to truly mismatch."""
        pass

    @property
    def match_threshold(self) -> float:
        """Similarity score (ranged 0..1) above which two references are considered to truly match one another."""
        pass

    def __call__(self, left: EntityReference, right: EntityReference) -> float:
        """Return a similarity score between ``left`` and ``right``.

        :param left: an ``EntityReference`` instance
        :param right: an ``EntityReference`` instance

        :return: a ``float`` value ranged between 0 and 1 which represents the
            probability that ``left`` matches ``right``.
        """
        raise NotImplementedError()



from dataclasses import dataclass
from typing import Hashable, Iterable, Sized, Protocol, Callable, Any

from matchescu.typing._data import Record


@dataclass(frozen=True, eq=True)
class EntityReferenceIdentifier:
    """Identifies an entity reference.

    Attributes:
        :label Hashable: unique label identifying an entity reference within a collection
        :source str: a string describing where the entity reference originated
    """

    label: Hashable
    source: str


class EntityReference(Record, Protocol):
    """An entity reference instance allows accessing data by name or index.

    Attributes:
        :id EntityReferenceIdentifier: identifies the entity reference
    """

    id: EntityReferenceIdentifier

    def as_dict(self) -> dict[str, Any]:
        """Return a dictionary containing the reference's attribute names and values."""
        pass


class EntityProfile(Iterable[EntityReference], Sized, Protocol):
    """Entity profiles are iterable sequences of a finite number of entity
    references.

    Concrete entity profiles have different representations according to the
    entity resolution model being used.
    """


EntityReferenceIdFactory = Callable[[Iterable[Record]], EntityReferenceIdentifier]



from collections.abc import Sequence, Iterable
from typing import Sized, Protocol, Union, Any, TypeVar, Callable, Generator


class Record(Sized, Iterable, Protocol):
    """A `protocol <https://peps.python.org/pep-0544/>`_ for data records.

    A record is information structured using attributes. A record has a length
    (or size), it can be iterated over so that we can browse all of its
    attributes and each attribute may be accessed using a name or an integer
    index.
    """

    def __getitem__(self, item: Union[str, int]) -> Any:
        """Record values may be accessed by name or index."""


# type variable for generic types that use records
TRecord = TypeVar("TRecord", bound=Record)

# traits take in one or more records and return a single record
Trait = Callable[[Iterable[TRecord]], TRecord]


class DataSource(Iterable[TRecord], Sized, Protocol):
    """A data source is an iterable sequence of relatively similar items.

    Data sources have a size or can at least estimate it. Each data source has a
    name.

    Attributes
    ----------
    :name str: name of the data source
    :traits Sequence[Trait]: entity reference extraction traits specific to this
        data source.
    """

    name: str
    traits: Sequence[Trait]


# Record samplers retrieve a finite number of records at a time from a data source.
RecordSampler = Callable[[DataSource], Generator[list[TRecord], None, None]]
